"""Helper functions"""


from bareasgi import Application
import bareasgi_jinja2
import jinja2
import pkg_resources


def add_swagger_ui(app: Application) -> None:
    """Add the Swagger UI

    Args:
        app (Application): The bareASGI application
    """
    templates = pkg_resources.resource_filename("bareasgi_rest", "templates")

    env = jinja2.Environment(
        loader=jinja2.FileSystemLoader(templates),
        autoescape=jinja2.select_autoescape(['html', 'xml']),
        enable_async=True
    )
    bareasgi_jinja2.add_jinja2(app, env)



"""Utility functions"""

from datetime import datetime
from decimal import Decimal
from inspect import Parameter, Signature
import json
import re
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    cast
)

from inflection import camelize, underscore

T = TypeVar('T')  # pylint: disable=invalid-name


def _rename_dict(dct: Dict[str, Any], rename: Callable[[str], str]) -> Dict[str, Any]:
    return {
        rename(name): _rename_object(obj, rename)
        for name, obj in dct.items()
    }


def _rename_list(lst: List[Any], rename: Callable[[str], str]) -> List[Any]:
    return [_rename_object(obj, rename) for obj in lst]


def _rename_object(obj: T, rename: Callable[[str], str]) -> T:
    if isinstance(obj, dict):
        return cast(T, _rename_dict(cast(Dict[str, Any], obj), rename))
    elif isinstance(obj, list):
        return cast(T, _rename_list(cast(List[Any], obj), rename))
    else:
        return obj


def _camelize(text: str) -> str:
    return camelize(text, uppercase_first_letter=False)


def camelize_object(obj: T) -> T:
    """Convert any dictionary keys in the object to camelcase

    :param obj: The object
    :type obj: T
    :return: The object with keys converted to camelcase
    :rtype: T
    """
    return _rename_object(obj, _camelize)


def underscore_object(obj: T) -> T:
    """Convert any dictionary keys in the object to underscore

    :param obj: The object
    :type obj: T
    :return: The object with keys converted to underscore
    :rtype: T
    """
    return _rename_object(obj, underscore)


class NullIter(Generic[T]):
    """An iterator containing no items"""

    def __aiter__(self):
        return self

    async def __anext__(self) -> T:
        raise StopAsyncIteration


def _is_supported_optional(annotation) -> bool:
    return (
        annotation is Optional[str] or
        annotation is Optional[int] or
        annotation is Optional[float] or
        annotation is Optional[Decimal] or
        annotation is Optional[datetime]
    )


def _coerce_builtin(value: Any, builtin_type: Type) -> Any:
    if isinstance(value, builtin_type):
        return value
    elif isinstance(value, str):
        if builtin_type is str:
            return value
        elif builtin_type is int:
            return int(value)
        elif builtin_type is float:
            return float(value)
        elif builtin_type is Decimal:
            return Decimal(value)
        elif builtin_type is datetime:
            return datetime.fromisoformat(value[:-1])
        else:
            raise TypeError(f'Unhandled type {builtin_type}')
    else:
        raise RuntimeError(f'Unable to coerce value {value}')


def _coerce(value: Any, annotation: Any) -> Any:
    if type(annotation) is type:  # pylint: disable=unidiomatic-typecheck
        single_value = value[0] if isinstance(value, list) else value
        return _coerce_builtin(single_value, annotation)

    if annotation is Optional[str]:
        return None if not value else _coerce(value, str)
    elif annotation is Optional[int]:
        return None if not value else _coerce(value, int)
    elif annotation is Optional[float]:
        return None if not value else _coerce(value, float)
    elif annotation is Optional[Decimal]:
        return None if not value else _coerce(value, Decimal)
    elif annotation is Optional[datetime]:
        return None if not value else _coerce(value, datetime)
    elif annotation is List[str]:
        contained_type: Any = str
    elif annotation is List[int]:
        contained_type = int
    elif annotation is List[float]:
        contained_type = float
    elif annotation is List[Decimal]:
        contained_type = Decimal
    elif annotation is List[datetime]:
        contained_type = datetime
    else:
        raise TypeError
    return [_coerce(item, contained_type) for item in value]


def make_args(
        sig: Signature,
        matches: Dict[str, str],
        query: Dict[str, Any],
        body: Dict[str, Any]
) -> Tuple[Tuple[Any, ...], Dict[str, Any]]:
    """Make args and kwargs for the given signature from the route matches,
    query args and body.

    :param sig: The function signature
    :type sig: Signature
    :param matches: The route matches
    :type matches: Dict[str, str]
    :param query: A dictionary built from the query string
    :type query: Dict[str, Any]
    :param body: A dictionary build from the body
    :type body: Dict[str, Any]
    :raises KeyError: If a parameter was not found
    :return: A tuple for *args and **kwargs
    :rtype: Tuple[Tuple[Any, ...], Dict[str, Any]]
    """
    kwargs: Dict[str, Any] = {}
    args: List[Any] = []

    for param in sig.parameters.values():
        name = camelize(param.name, uppercase_first_letter=False)
        if name in matches:
            value = _coerce(matches[name], param.annotation)
        elif name in query:
            value = _coerce(query[name], param.annotation)
        elif name in body:
            value = _coerce(body[name], param.annotation)
        elif _is_supported_optional(param.annotation):
            continue
        else:
            raise KeyError

        if param.kind == Parameter.POSITIONAL_ONLY or param.kind == Parameter.POSITIONAL_OR_KEYWORD:
            args.append(value)
        else:
            # Use the non-camelcased name
            kwargs[param.name] = value

    bound_args = sig.bind(*args, **kwargs)
    bound_args.apply_defaults()

    return bound_args.args, bound_args.kwargs


DATETIME_ZULU_REGEX = re.compile(r'^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}Z$')


def as_datetime(dct):
    """Convert datetime like strings"""
    for key, value in dct.items():
        if isinstance(value, str) and DATETIME_ZULU_REGEX.match(value):
            dct[key] = datetime.fromisoformat(value[:-1])
    return dct


class JSONEncoderEx(json.JSONEncoder):
    """Encode json"""

    def default(self, obj):  # pylint: disable=method-hidden,arguments-differ
        if isinstance(obj, datetime):
            return obj.isoformat() + ('Z' if not obj.tzinfo else '')
        elif isinstance(obj, Decimal):
            return float(
                str(obj.quantize(Decimal(1))
                    if obj == obj.to_integral() else
                    obj.normalize())
            )
        else:
            return super(JSONEncoderEx, self).default(obj)



"""Configuration"""

from typing import List, NamedTuple, Optional


class SwaggerOauth2Config(NamedTuple):
    """Swagger Oauth2 config

    Args:
        client_id (str): The client id
        realm (str): The realm
        app_name (str): The application name
        redirect_url (str): The redirect url
    """
    client_id: str
    realm: str
    app_name: str
    redirect_url: str


class SwaggerConfig(NamedTuple):
    """The Swagger UI Configuration

    Args:
        title (str, optional): The page title, defaults to "The REST API"
        validator_url (Optional[str], optional): The url of the validator,
            defaults to None
        supported_submit_methods (Optional[List[str]], optional): The supported
            submit methods, defaults to None
        display_operation_id (bool, optional): Whether operation ids should be
            displayed, defaults to False
        display_request_duration (bool, optional): Whether the duration of the
            request should be displayed, defaults to False
        doc_expansion (str, optional): How the methods should be displayed,
            defaults to "list"
        oauth2 (Optional[SwaggerOauth2Config], optional): Optional oauth2
            config, defaults to None

    """
    title: str = "The REST API"
    validator_url: Optional[str] = None
    supported_submit_methods: Optional[List[str]] = None
    display_operation_id: bool = False
    display_request_duration: bool = False
    doc_expansion: str = "list"
    oauth2: Optional[SwaggerOauth2Config] = None



"""bareASGI-rest"""

from .rest_router import RestHttpRouter
from .helpers import add_swagger_ui
from .config import SwaggerConfig, SwaggerOauth2Config

__all__ = [
    "RestHttpRouter",
    "add_swagger_ui",
    "SwaggerConfig",
    "SwaggerOauth2Config"
]



"""Utility functions"""

from datetime import datetime
from decimal import Decimal
import inspect
from inspect import Parameter, Signature
from typing import (
    AbstractSet,
    Any,
    Dict,
    List,
    Optional,
    Set,
    Tuple,
    cast
)

from bareasgi.basic_router.path_definition import PathDefinition
import docstring_parser
from docstring_parser import Docstring, DocstringParam
from inflection import underscore, camelize
import typing_inspect


def make_swagger_path(path_definition: PathDefinition) -> str:
    """Make a path compatible with swagger"""
    swagger_path = '/' + '/'.join(
        '{' + segment.name + '}' if segment.is_variable else segment.name
        for segment in path_definition.segments
    )
    if path_definition.ends_with_slash:
        swagger_path += '/'
    return swagger_path


TYPE_DEFINITIONS = {
    str: {
        'is_required': True,
        'is_list': False,
        'type': 'string',
        'format': None,
    },
    int: {
        'is_required': True,
        'is_list': False,
        'type': 'integer',
        'format': None,
    },
    float: {
        'is_required': True,
        'type': 'number',
        'format': None,
    },
    Decimal: {
        'is_required': True,
        'is_list': False,
        'type': 'number',
        'format': None,
    },
    datetime: {
        'is_required': True,
        'is_list': False,
        'type': 'string',
        'format': 'date',
    },
    Optional[str]: {
        'is_required': False,
        'is_list': False,
        'type': 'string',
        'format': None,
    },
    Optional[int]: {
        'is_required': False,
        'is_list': False,
        'type': 'integer',
        'format': None,
    },
    Optional[float]: {
        'is_required': False,
        'is_list': False,
        'type': 'number',
        'format': None,
    },
    Optional[Decimal]: {
        'is_required': False,
        'is_list': False,
        'type': 'number',
        'format': None,
    },
    Optional[datetime]: {
        'is_required': False,
        'is_list': False,
        'type': 'string',
        'format': 'date',
    },
    List[str]: {
        'is_required': False,
        'is_list': True,
        'type': 'string',
        'format': None,
    },
    List[int]: {
        'is_required': False,
        'is_list': True,
        'type': 'integer',
        'format': None,
    },
    List[float]: {
        'is_required': False,
        'is_list': True,
        'type': 'string',
        'format': None,
    },
    List[Decimal]: {
        'is_required': False,
        'is_list': True,
        'type': 'number',
        'format': None,
    },
    List[datetime]: {
        'is_required': False,
        'is_list': True,
        'type': 'string',
        'format': 'date',
    }
}


def _check_is_required(
        param: Parameter,
        docstring_param: DocstringParam
) -> bool:
    type_def = TYPE_DEFINITIONS[param.annotation]
    is_required: bool = cast(bool, type_def['is_required']) or (
        docstring_param is not None and not docstring_param.is_optional
    )
    return is_required


def _make_swagger_parameter(
        source: str,
        param: Parameter,
        collection_format: str,
        docstring_param: Optional[DocstringParam]
) -> Dict[str, Any]:
    type_def = TYPE_DEFINITIONS[param.annotation]
    is_required = _check_is_required(param, docstring_param)

    parameter = {
        'name': camelize(param.name, False),
        'type': 'array' if type_def['is_list'] else type_def['type']
    }

    if type_def['format'] is not None:
        parameter['format'] = type_def['format']

    if type_def['is_list']:
        parameter['collectionFormat'] = collection_format
        items = {
            'type': type_def['type']
        }
        if type_def['format'] is not None:
            items['format'] = type_def['format']
        parameter['items'] = items

    if source != 'body':
        parameter['in'] = source
        parameter['required'] = is_required

    if param.default != Parameter.empty:
        parameter['default'] = param.default

    if docstring_param is not None and docstring_param.description:
        parameter['description'] = docstring_param.description
    return parameter


def _make_swagger_schema(
        params: List[Tuple[Parameter, DocstringParam]],
        collection_format: str
) -> Dict[str, Any]:
    return {
        'type': 'object',
        'required': [
            param.name
            for param, docstring_param in params
            if _check_is_required(param, docstring_param)
        ],
        'properties': {
            param.name: _make_swagger_parameter(
                'body',
                param,
                collection_format,
                docstring_param
            )
            for param, docstring_param in params
        }
    }


def _find_docstring_param(
        name: str,
        docstring: Docstring
) -> Optional[DocstringParam]:
    for param in docstring.params:
        if param.arg_name == name:
            return param
    return None


def _make_swagger_parameters_inline(
        source: str,
        sig: Signature,
        path_variables: AbstractSet[str],
        docstring: Docstring,
        collection_format: str
) -> List[Dict[str, Any]]:
    """Make inline paramters for query or form"""
    parameters: List[Dict[str, Any]] = []
    for param in sig.parameters.values():
        if param.name in path_variables:
            continue
        docstring_param = _find_docstring_param(param.name, docstring)
        parameter = _make_swagger_parameter(
            source,
            param,
            collection_format,
            docstring_param
        )
        parameters.append(parameter)
    return parameters


def make_swagger_parameters(
        method: str,
        accept: bytes,
        path_definition: PathDefinition,
        sig: Signature,
        docstring: Docstring,
        collection_format: str
) -> List[Dict[str, Any]]:
    """Make the swagger parameters"""
    parameters: List[Dict[str, Any]] = []
    path_variables: Set[str] = set()
    for segment in path_definition.segments:
        if segment.is_variable:
            path_variable = underscore(segment.name)
            path_variables.add(path_variable)
            param = sig.parameters[path_variable]
            docstring_param = _find_docstring_param(param.name, docstring)
            parameter = _make_swagger_parameter(
                'path',
                param,
                collection_format,
                docstring_param
            )
            parameters.append(parameter)

    if method == 'GET':
        parameters.extend(
            _make_swagger_parameters_inline(
                'query',
                sig,
                path_variables,
                docstring,
                collection_format
            )
        )
    elif accept in (b'application/x-www-form-urlencoded', b'multipart/form-data'):
        parameters.extend(
            _make_swagger_parameters_inline(
                'formData',
                sig,
                path_variables,
                docstring,
                collection_format
            )
        )
    else:
        params = [
            (param, _find_docstring_param(param.name, docstring))
            for param in sig.parameters.values()
            if param.name not in path_variables
        ]
        schema = _make_swagger_schema(
            params,
            collection_format
        )
        parameters.append({
            'in': 'body',
            'name': 'schema',
            'description': 'The body schema',
            'schema': schema
        })

    return parameters


def gather_error_responses(docstring: Docstring) -> Dict[int, Any]:
    """Gather error responses"""
    responses: Dict[int, Any] = {}
    for raises in docstring.raises:
        if raises.type_name != 'HTTPError':
            continue
        first, sep, rest = raises.description.partition(',')
        if not sep:
            continue
        try:
            error_code = int(first.strip())
            responses[error_code] = {
                'description': rest.strip()
            }
        except:  # pylint: disable=bare-except
            continue
    return responses


def _typeddict_schema(
        schema_type: str,
        annotations: Dict[str, Any],
        docstring: Optional[Docstring]
) -> Dict[str, Any]:
    properties: Dict[str, Any] = {}
    for name, value in annotations.items():
        prop: Dict[str, Any] = {}
        docstring_param = _find_docstring_param(name, docstring)
        if docstring_param is not None:
            prop['description'] = docstring_param.description
        type_def = TYPE_DEFINITIONS.get(value)
        if type_def is not None:
            prop['type'] = type_def['type']
            if type_def['format'] is not None:
                prop['format'] = type_def['format']
        camelcase_name = camelize(name, False)
        prop['name'] = camelcase_name
        properties[camelcase_name] = prop

    if schema_type == 'object':
        return {
            'type': 'object',
            'properties': properties
        }
    else:
        return {
            'type': 'array',
            'items': {
                'type': 'object',
                'properties': properties
            }
        }


def make_swagger_response_schema(
        sig: Signature
) -> Optional[Dict[str, Any]]:
    """Make the swagger response schama"""
    if sig.return_annotation is None:
        return None

    origin = typing_inspect.get_origin(sig.return_annotation)
    if not origin and issubclass(sig.return_annotation, dict):
        # could be a typed dict
        annotations = getattr(sig.return_annotation, '__annotations__', None)
        if isinstance(annotations, dict):
            docstring = docstring_parser.parse(
                inspect.getdoc(sig.return_annotation)
            )
            return _typeddict_schema('object', annotations, docstring)
        else:
            return None
    elif origin and origin is list:
        # could be a list of typed dicts
        args = typing_inspect.get_args(sig.return_annotation)
        if len(args) != 1:
            return None
        nested_type = args[0]
        nested_origin = typing_inspect.get_origin(nested_type)
        if nested_origin and nested_origin is dict:
            # List[Dict]
            return None
        elif not nested_origin and issubclass(nested_type, dict):
            # A TypedDict
            annotations = getattr(nested_type, '__annotations__', None)
            if isinstance(annotations, dict):
                # List[TypedDict]
                docstring = docstring_parser.parse(inspect.getdoc(nested_type))
                return _typeddict_schema('array', annotations, docstring)
            else:
                # List[Dict]
                return None
        else:
            # List
            return None
    elif origin and origin is dict:
        # A Dict
        return None

    # Something else
    return None



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['bareasgi_rest']

package_data = \
{'': ['*'], 'bareasgi_rest': ['templates/*']}

install_requires = \
['bareasgi-jinja2>=3.1,<4.0',
 'bareasgi>=3.5,<4.0',
 'docstring-parser>=0.5,<0.6',
 'inflection>=0.3.1,<0.4.0',
 'typing-extensions>=3.7,<4.0',
 'typing-inspect>=0.5.0,<0.6.0']

setup_kwargs = {
    'name': 'bareasgi-rest',
    'version': '3.0.0rc1',
    'description': 'REST support for bareASGI',
    'long_description': '# bareASGI-rest\n\n\nThis package provides enhanced support for writing REST\nAPIs with [bareASGI](https://bareasgi.com).\n\nIt includes:\n\n* A router to simplify the creation of REST APIs,\n* A swagger API endpoint\n\nThis is a Python 3.7+ package, and is currently pre-release.\n\n## Installation\n\nThe package can be install from pypi.\n\nIt is currently pre-release so you will need the --pre flag.\n\n```bash\n$ pip install --pre bareASGI-rest\n```\n\nAn ASGI server will be required to run the code. The examples below use\n[uvicorn](https://www.uvicorn.org/).\n\n```bash\n$ pip install uvicorn\n```\n\n## Usage\n\nThe router provided by this package maps the arguments and\ntypes of request handlers.\n\nWe will create a mock book repository.\n\n### Creating typed dictionaries\n\nHere is the type of a book. We use `TypedDict` to allow automatic type discovery\n\n```python\nfrom datetime import datetime\ntry:\n    # Available in 3.8\n    from typing import TypedDict  # type:ignore\nexcept:\n    # Available in 3.7\n    from typing_extensions import TypedDict\n\nclass Book(TypedDict):\n    """A Book\n\n    Args:\n        book_id (int): The book id\n        title (str): The title\n        author (str): The author\n        published (datetime): The publication date\n    """\n    book_id: int\n    title: str\n    author: str\n    published: datetime\n```\n\nNote: the docstring will be used to provide documentation for swagger.\n\n### Creating the API\n\nNow we can build the API.\n\n```python\nfrom Typing import Dict, List\nfrom urllib.error import HTTPError\n\n\nBOOKS: Dict[int, Book]\nNEXT_ID: int = 0\n\nasync def get_books() -> List[Book]:\n    """Get all the books.\n\n    This method gets all the books in the shop.\n\n    Returns:\n        List[Book]: All the books\n    """\n    return list(BOOKS.values())\n\n\nasync def get_book(book_id: int) -> Book:\n    """Get a book for a given id\n\n    Args:\n        book_id (int): The id of the book\n\n    Raises:\n        HTTPError: 404, when a book is not found\n\n    Returns:\n        Book: The book\n    """\n\n    if book_id not in BOOKS:\n        raise HTTPError(None, 404, None, None, None)\n\n    return BOOKS[book_id]\n\n\nasync def create_book(\n        author: str,\n        title: str,\n        published: datetime\n) -> int:\n    """Add a book\n\n    Args:\n        author (str): The author\n        title (str): The title\n        published (datetime): The publication date\n\n    Returns:\n        int: The id of the new book\n    """\n    NEXT_ID += 1\n    BOOKS[NEXT_ID] = Book(\n        book_id=NEXT_ID,\n        title=title,\n        author=author,\n        published=published\n    )\n    return NEXT_ID\n\n\nasync def update_book(\n        book_id: int,\n        author: str,\n        title: str,\n        published: datetime\n) -> None:\n    """Update a book\n\n    Args:\n        book_id (int): The id of the book to update\n        author (str): The new author\n        title (str): The title\n        published (datetime): The publication date\n\n    Raises:\n        HTTPError: 404, when a book is not found\n    """\n    if book_id not in BOOKS:\n        raise HTTPError(None, 404, None, None, None)\n    BOOKS[book_id][\'title\'] = title\n    BOOKS[book_id][\'author\'] = author\n    BOOKS[book_id][\'published\'] = published\n```\n\nWe can see that errors are handler by raising HTTPError\nfrom the `urllib.errors` standard library package. A convention has been applied such that the status code MUST\nappear before the message, separated by a comma.\n\n### Adding support for the REST router\n\nNow we must create our application and add support for the router.\n\n```python\nfrom bareasgi_rest import RestHTTPRouter, add_swagger_ui\n\n\nrest_router = RestHttpRouter(\n    None,\n    title="Books",\n    version="1",\n    description="A book api",\n    base_path=\'/api/1\',\n    tags=[\n        {\n            \'name\': \'Books\',\n            \'description\': \'The book store API\'\n        }\n    ]\n)\napp = Application(http_router=rest_router)\nadd_swagger_ui(app)\n```\n\nNote the `base-path` argument can be used to prefix all\npaths.\n\nThe `RestHttpRouter` is a subclass of the basic router, so\nall those methods are also available.\n\n### Creating the routes\n\nNow we can create the routes:\n\n```python\ntags = [\'Books\']\nrouter.add_rest({\'GET\'}, \'/books\', get_books,tags=tags)\nrouter.add_rest({\'GET\'}, \'/books/{bookId:int}\', get_book, tags=tags)\nrouter.add_rest({\'POST\'}, \'/books\', create_book, tags=tags, status_code=201)\nrouter.add_rest({\'PUT\'}, \'/books/{bookId:int}\', update_book, tags=tags, status_code=204)\n```\n\nFirst we should note that the paths with be prefix with the\n`base_url_path` provided to the router.\n\nReferring back to the implementation of `get_book` we can\nsee  that the camel-case path variable `bookId` has been\nmapped to the snake-case `book_id` parameter. The JSON object provided in the body of the `create_book` will\nsimilarly map camel-cased properties to the snake-cased\nfunction parameters.\n\nWe can also see how the status codes have been overridden\nfor the `POST` and `PUT` endpoints, and all the routes\nhave the "Books" tag for grouping in the UI.\n\n### Serving the API\n\nFinally we can serve the API:\n\n```python\nuvicorn.run(app, port=9009)\n```\n\nBrowsing to http://localhost/api/1/swagger we should see:\n\n![Top Level](screenshot1.png)\n\nWhen we expand `GET /books/{bookId}` we can see all the\ninformation provided in the docstring and typing has been\npassed through to the swagger UI.\n\n![GET /books/{bookId}](screenshot2.png)\n\n## Thanks\n\nThanks to [rr-](https://github.com/rr-) and contributors\nfor the excellent\n[docstring-parser](https://github.com/rr-/docstring_parser)\npackage.\n',
    'author': 'Rob Blackbourn',
    'author_email': 'rob.blackbourn@gmail.com',
    'url': 'https://github.com/rob-blackbourn/bareASGI-rest',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""A router for REST APIs

Attributes:
    DEFAULT_SWAGGER_BASE_URL (str): The default swagger CDN url. The currently
        supported version is 3.4.0
    DEFAULT_TYPEFACE_URL (str): The typeface url to use.
    DEFAULT_CONSUMES
"""

from cgi import parse_multipart
import io
import inspect
import logging
import json
from typing import (
    AbstractSet,
    Any,
    AsyncIterator,
    Awaitable,
    Callable,
    Dict,
    List,
    Mapping,
    Optional,
    cast
)
from urllib.parse import parse_qs

from bareasgi import text_reader, text_writer
from bareasgi.basic_router.http_router import BasicHttpRouter, PathDefinition
from baretypes import (
    RouteMatches,
    Scope,
    Info,
    Content,
    HttpResponse,
    Headers
)
import bareutils.header as header
import bareasgi_jinja2
import docstring_parser

from .utils import make_args, JSONEncoderEx, as_datetime, camelize_object
from .swagger import (
    make_swagger_path,
    make_swagger_parameters,
    gather_error_responses,
    make_swagger_response_schema
)
from .config import SwaggerConfig

LOGGER = logging.getLogger(__name__)

DEFAULT_SWAGGER_BASE_URL = "https://cdnjs.cloudflare.com/ajax/libs/swagger-ui/3.4.0"
DEFAULT_TYPEFACE_URL = "https://fonts.googleapis.com/css?family=Roboto:300,400,500,700&display=swap"

APPLICATION_JSON = b'application/json'


def _to_json(obj: Any) -> str:
    return json.dumps(obj, cls=JSONEncoderEx)


def _from_json(text: str, _media_type: bytes, _params: Dict[bytes, bytes]) -> Any:
    return json.loads(text, object_hook=as_datetime)


def _from_query_string(text: str, _media_type: bytes, _params: Dict[bytes, bytes]) -> Any:
    return parse_qs(text)


def _from_form_data(text: str, _media_type: bytes, params: Dict[bytes, bytes]) -> Any:
    if b'boundary' not in params:
        raise RuntimeError('Required "boundary" parameter missing')
    pdict = {
        name.decode(): value
        for name, value in params.items()
    }
    return parse_multipart(io.StringIO(text), pdict)


Deserializer = Callable[[str, bytes, Dict[bytes, bytes]], Any]
DictConsumes = Dict[bytes, Deserializer]
DEFAULT_CONSUMES: DictConsumes = {
    b'application/json': _from_json,
    b'*/*': _from_json,
    b'application/x-www-form-urlencoded': _from_query_string,
    b'multipart/form-data': _from_form_data
}

Serializer = Callable[[Any], str]
DictProduces = Dict[bytes, Serializer]
DEFAULT_PRODUCES: DictProduces = {
    b'application/json': _to_json,
    b'*/*': _to_json
}

RestCallback = Callable[..., Awaitable[Any]]

DEFAULT_COLLECTION_FORMAT = 'multi'
DEFAULT_RESPONSES = {
    200: {
        'description': 'OK'
    }
}

DEFAULT_NOT_FOUND_RESPONSE: HttpResponse = (
    404,
    [(b'content-type', b'text/plain')],
    text_writer('Not Found'),
    None
)


class RestHttpRouter(BasicHttpRouter):
    """A REST router"""

    def __init__(
            self,
            not_found_response: HttpResponse,
            *,
            title: str,
            version: str,
            description: Optional[str] = None,
            base_path: str = '',
            consumes: DictConsumes = DEFAULT_CONSUMES,
            produces: DictProduces = DEFAULT_PRODUCES,
            tags: Optional[List[Mapping[str, Any]]] = None,
            swagger_base_url: str = DEFAULT_SWAGGER_BASE_URL,
            typeface_url: str = DEFAULT_TYPEFACE_URL,
            config: Optional[SwaggerConfig] = None
    ) -> None:
        """Initialise the REST router

        Args:
            not_found_response (HttpResponse): The response sent when a route is
                not found
            title (str): The title of the swagger documentation
            version (str): The version of the exposed API
            description (Optional[str], optional): The API description. Defaults
                to None.
            base_path (str, optional): The base path of the API. Defaults to ''.
            consumes (DictConsumes, optional): A map of media types and
                deserializers. Defaults to DEFAULT_CONSUMES.
            produces (DictProduces, optional): A map of media types and
                serializers. Defaults to DEFAULT_PRODUCES.
            tags (Optional[List[Mapping[str, Any]]], optional): The available
                tags. Defaults to None.
            swagger_base_url (Optional[str], optional): The base url for the
                swagger CDN. Defaults to DEFAULT_SWAGGER_BASE_URL.
            typeface_url (Optional[str], optional): The base url for the
                typeface. Defaults to DEFAULT_TYPEFACE_URL.
            config (Optional[SwaggerConfig], optional): The swagger
                configuration. Defaults to None.
        """
        super().__init__(not_found_response or DEFAULT_NOT_FOUND_RESPONSE)
        self.title = title
        self.version = version
        self.description = description
        self.consumes = consumes
        self.produces = produces
        self.base_path = base_path
        self.swagger_base_url = swagger_base_url
        self.typeface_url = typeface_url
        self.config = config or SwaggerConfig()

        self.accepts: Dict[str, Dict[PathDefinition, bytes]] = {}
        self.collection_formats: Dict[str, Dict[PathDefinition, str]] = {}

        self.add({'GET'}, base_path + '/swagger.json', self.swagger_json)
        self.add({'GET'}, base_path + '/swagger', self.swagger_ui)

        self.swagger_dict: Dict[str, Any] = {
            'swagger': '2.0',
            'basePath': self.base_path,
            'info': {
                'title': self.title,
                'version': self.version,
                'description': self.description
            },
            'produces': [name.decode() for name in self.produces.keys()],
            'consumes': [name.decode() for name in self.consumes.keys()],
            "paths": {},
        }
        if tags:
            self.swagger_dict['tags'] = tags

    def add_rest(
            self,
            methods: AbstractSet[str],
            path: str,
            callback: RestCallback,
            *,
            accept: bytes = APPLICATION_JSON,
            content_type: bytes = APPLICATION_JSON,
            collection_format: str = DEFAULT_COLLECTION_FORMAT,
            tags: Optional[List[str]] = None,
            status_code: int = 200,
            status_description: str = 'OK'
    ) -> None:
        """Register a callback to a method and path

        Args:
            methods (AbstractSet[str]): The set of methods
            path (str): The path
            callback (RestCallback): The callback
            accept (bytes, optional): The accept media type. Defaults to
                APPLICATION_JSON.
            content_type (bytes, optional): The content media type. Defaults
                to APPLICATION_JSON.
            collection_format (str, optional): The format of repeated values.
                Defaults to DEFAULT_COLLECTION_FORMAT.
            tags (Optional[List[str]], optional): A list of tags. Defaults to
                None.
            status_code (int, optional): The ok status code. Defaults to 200.
            status_description (str, optional): The ok status message. Defaults
                to 'OK'.
        """
        LOGGER.debug('Adding route for %s on "%s"', methods, path)

        for method in methods:
            self._add_method(
                method,
                path,
                callback,
                content_type,
                status_code
            )
            self._add_swagger_path(
                method,
                path,
                callback,
                accept,
                content_type,
                collection_format,
                tags,
                status_code,
                status_description
            )

    def _add_method(
            self,
            method: str,
            path: str,
            callback: RestCallback,
            content_type: bytes,
            status_code: int
    ) -> None:
        sig = inspect.signature(callback)
        path_definition = PathDefinition(self.base_path + path)

        async def rest_callback(
                scope: Scope,
                _info: Info,
                matches: RouteMatches,
                content: Content
        ) -> HttpResponse:

            query_args = parse_qs(scope['query_string'].decode())
            body_args = await self._get_body_args(method, scope['headers'], content)

            args, kwargs = make_args(
                sig,
                matches,
                query_args,
                body_args
            )

            body = await callback(*args, **kwargs)

            accept = header.accept(scope['headers'])
            writer = self._make_writer(body, accept)
            headers = [
                (b'content-type', content_type)
            ]
            return status_code, headers, writer

        self.add_route(method, path_definition, rest_callback)

    def _add_swagger_path(
            self,
            method: str,
            path: str,
            callback: RestCallback,
            accept: bytes,
            content_type: bytes,
            collection_format: str,
            tags: Optional[List[str]],
            status_code: int,
            status_description: str
    ):
        path_definition = PathDefinition(path)
        swagger_path = make_swagger_path(path_definition)
        sig = inspect.signature(callback)
        docstring = docstring_parser.parse(inspect.getdoc(callback))
        params = make_swagger_parameters(
            method,
            accept,
            path_definition,
            sig,
            docstring,
            collection_format
        )

        response: Dict[str, Any] = {
            'description': status_description
        }

        response_schema = make_swagger_response_schema(sig)
        if response_schema is not None:
            response['schema'] = response_schema

        entry = {
            'parameters': params,
            'produces': [content_type.decode()],
            'consumes': [accept.decode()],
            'responses': {
                status_code: response
            }
        }

        if docstring:
            if docstring.short_description:
                entry['summary'] = docstring.short_description
            if docstring.long_description:
                entry['description'] = docstring.long_description
            error_responses = gather_error_responses(docstring)
            responses = cast(Dict[int, Dict[str, Any]], entry['responses'])
            responses.update(error_responses)

        if tags:
            entry['tags'] = tags

        paths: Dict[str, Any] = self.swagger_dict['paths']
        current_path: Dict[str, Any] = paths.setdefault(swagger_path, {})
        current_path[method.lower()] = entry

    def _make_writer(
            self,
            data: Optional[Any],
            accept: Optional[Mapping[bytes, float]]
    ) -> Optional[AsyncIterator[bytes]]:
        if data is None:
            return None
        # TODO: This is rubbish
        if not accept:
            serializer = self.produces[APPLICATION_JSON]
        else:
            for media_type in accept.keys():
                if media_type in self.produces:
                    serializer = self.produces[media_type]
                    break
            else:
                serializer = self.produces[APPLICATION_JSON]
        if not serializer:
            raise RuntimeError
        text = serializer(camelize_object(data))
        return text_writer(text)

    async def _get_body_args(
            self,
            method: str,
            headers: Headers,
            content: Content
    ) -> Any:
        if method in {'GET'}:
            return True, {}

        media_type, params = header.content_type(
            headers) or (b'', cast(Dict[bytes, Any], {}))
        deserializer = self.consumes.get(media_type)
        if deserializer is None:
            raise RuntimeError('No deserializer')
        body = await text_reader(content)
        return deserializer(body, media_type, params)

    async def swagger_json(
            self,
            _scope: Scope,
            _info: Info,
            _matches: RouteMatches,
            _content: Content
    ) -> HttpResponse:
        """The swagger JSON request handler"""
        spec = json.dumps(self.swagger_dict)
        return 200, [(b'content-type', b'application/json')], text_writer(spec)

    @bareasgi_jinja2.template('swagger.html')
    async def swagger_ui(
            self,
            _scope: Scope,
            _info: Info,
            _matches: RouteMatches,
            _content: Content
    ) -> Dict[str, Any]:
        """The swagger view request handler"""
        return {
            "title": self.title,
            "specs_url": "/api/1/swagger.json",
            'swagger_base_url': self.swagger_base_url,
            'typeface_url': self.typeface_url,
            "config": self.config
        }
